@@ -1,6 +1,6 @@
 /*	MikMod sound library
-	(c) 1998, 1999 Miodrag Vallat and others - see file AUTHORS for
-	complete list.
+	(c) 1998, 1999, 2000, 2001, 2002 Miodrag Vallat and others - see file
+	AUTHORS for complete list.
 
 	This library is free software; you can redistribute it and/or modify
 	it under the terms of the GNU Library General Public License as
@@ -20,7 +20,7 @@
 
 /*==============================================================================
 
-  $Id: load_xm.c,v 1.30 1999/06/28 03:00:05 miod Exp $
+  $Id: load_xm.c,v 1.1.1.2 2007/12/15 15:24:17 denis111 Exp $
 
   Fasttracker (XM) module loader
 
@@ -30,9 +30,21 @@
 #include "config.h"
 #endif
 
+#ifdef HAVE_UNISTD_H
+#include <unistd.h>
+#endif
+
+#include <stdio.h>
+#ifdef HAVE_MEMORY_H
+#include <memory.h>
+#endif
 #include <string.h>
 
-#include <mikmod_internals.h>
+#include "mikmod_internals.h"
+
+#ifdef SUNOS
+extern int fprintf(FILE *, const char *, ...);
+#endif
 
 /*========== Module structure */
 
@@ -142,8 +154,7 @@ BOOL XM_Init(void)
 
 void XM_Cleanup(void)
 {
-	if(mh) free(mh);
-	mh = NULL;
+	_mm_free(mh);
 }
 
 static int XM_ReadNote(XMNOTE* n)
@@ -209,10 +220,10 @@ static UBYTE* XM_Convert(XMNOTE* xmtrack,UWORD rows)
 				UniPTEffect(0xe,0xa0|(vol&0xf));
 				break;
 			case 0xa: /* set vibrato speed */
-				UniPTEffect(0x4,vol<<4);
+				UniEffect(UNI_XMEFFECT4,vol<<4);
 				break;
 			case 0xb: /* vibrato */
-				UniPTEffect(0x4,vol&0xf);
+				UniEffect(UNI_XMEFFECT4,vol&0xf);
 				break;
 			case 0xc: /* set panning */
 				UniPTEffect(0x8,vol<<4);
@@ -235,6 +246,9 @@ static UBYTE* XM_Convert(XMNOTE* xmtrack,UWORD rows)
 			case 0x4:
 				UniEffect(UNI_XMEFFECT4,dat);
 				break;
+			case 0x6:
+				UniEffect(UNI_XMEFFECT6,dat);
+				break;
 			case 0xa:
 				UniEffect(UNI_XMEFFECTA,dat);
 				break;
@@ -257,7 +271,7 @@ static UBYTE* XM_Convert(XMNOTE* xmtrack,UWORD rows)
 				}
 				break;
 			case 'G'-55: /* G - set global volume */
-				UniEffect(UNI_XMEFFECTG,dat>64?64:dat);
+				UniEffect(UNI_XMEFFECTG,dat>64?128:dat<<1);
 				break;
 			case 'H'-55: /* H - global volume slide */
 				UniEffect(UNI_XMEFFECTH,dat);
@@ -389,11 +403,49 @@ static BOOL LoadPatterns(BOOL dummypat)
 	return 1;
 }
 
+static void FixEnvelope(ENVPT *cur, int pts)
+{
+		int u, old, tmp;
+		ENVPT *prev;
+
+		/* Some broken XM editing program will only save the low byte
+		   of the position value. Try to compensate by adding the
+		   missing high byte. */
+
+		prev = cur++;
+		old = prev->pos;
+
+		for (u = 1; u < pts; u++, prev++, cur++) {
+			if (cur->pos < prev->pos) {
+				if (cur->pos < 0x100) {
+					if (cur->pos > old)	/* same hex century */
+							tmp = cur->pos + (prev->pos - old);
+					else
+							tmp = cur->pos | ((prev->pos + 0x100) & 0xff00);
+					old = cur->pos;
+					cur->pos = tmp;
+#ifdef MIKMOD_DEBUG
+					fprintf(stderr, "\rbroken envelope position(%d/%d), %d %d -> %d\n",
+					    u, pts, prev->pos, old, cur->pos);
+#endif
+				} else {
+#ifdef MIKMOD_DEBUG
+					/* different brokenness style... fix unknown */
+					fprintf(stderr, "\rbroken envelope position(%d/%d), %d %d\n",
+					    u, pts, old, cur->pos);
+#endif
+					old = cur->pos;
+				}
+			} else
+				old = cur->pos;
+		}
+}
+
 static BOOL LoadInstruments(void)
 {
-	int t,u;
+	int t,u, ck;
 	INSTRUMENT *d;
-	long next=0;
+	ULONG next=0;
 	UWORD wavcnt=0;
 
 	if(!AllocInstruments()) return 0;
@@ -408,6 +460,13 @@ static BOOL LoadInstruments(void)
 		headend     = _mm_ftell(modreader);
 		ih.size     = _mm_read_I_ULONG(modreader);
 		headend    += ih.size;
+		ck = _mm_ftell(modreader);
+		_mm_fseek(modreader,0,SEEK_END);
+		if ((headend<0) || (_mm_ftell(modreader)<headend) || (headend<ck)) {
+		_mm_fseek(modreader,ck,SEEK_SET);
+		break;
+		}
+		_mm_fseek(modreader,ck,SEEK_SET);
 		_mm_read_string(ih.name, 22, modreader);
 		ih.type     = _mm_read_UBYTE(modreader);
 		ih.numsmp   = _mm_read_I_UWORD(modreader);
@@ -441,10 +500,13 @@ static BOOL LoadInstruments(void)
 
 				/* read the remainder of the header
 				   (2 bytes for 1.03, 22 for 1.04) */
-				for(u=headend-_mm_ftell(modreader);u;u--) _mm_read_UBYTE(modreader);
+				if (headend>=_mm_ftell(modreader)) for(u=headend-_mm_ftell(modreader);u;u--) _mm_read_UBYTE(modreader);
 
-				/* #@!$&% fix for K_OSPACE.XM */
-				if(pth.volpts==32) pth.volpts=XMENVCNT/2;
+				/* we can't trust the envelope point count here, as some
+				   modules have incorrect values (K_OSPACE.XM reports 32 volume
+				   points, for example). */
+				if(pth.volpts>XMENVCNT/2) pth.volpts=XMENVCNT/2;
+				if(pth.panpts>XMENVCNT/2) pth.panpts=XMENVCNT/2;
 
 				if((_mm_eof(modreader))||(pth.volpts>XMENVCNT/2)||(pth.panpts>XMENVCNT/2)) {
 					if(nextwav) { free(nextwav);nextwav=NULL; }
@@ -456,34 +518,58 @@ static BOOL LoadInstruments(void)
 				for(u=0;u<XMNOTECNT;u++)
 					d->samplenumber[u]=pth.what[u]+of.numsmp;
 				d->volfade = pth.volfade;
-				memcpy(d->volenv,pth.volenv,XMENVCNT);
-				if(pth.volflg & 1)  d->volflg |= EF_ON;
-				if(pth.volflg & 2)  d->volflg |= EF_SUSTAIN;
-				if(pth.volflg & 4)  d->volflg |= EF_LOOP;
-				d->volsusbeg = d->volsusend = pth.volsus;
-				d->volbeg    = pth.volbeg;
-				d->volend    = pth.volend;
-				d->volpts    = pth.volpts;
 
-				/* scale volume envelope: */
-				for(p=0;p<XMENVCNT/2;p++)
-					d->volenv[p].val<<=2;
+#if defined __STDC__ || defined _MSC_VER
+#define XM_ProcessEnvelope(name) 										\
+				for (u = 0; u < (XMENVCNT >> 1); u++) {					\
+					d-> name##env[u].pos = pth. name##env[u << 1];		\
+					d-> name##env[u].val = pth. name##env[(u << 1)+ 1];	\
+				}														\
+				if (pth. name##flg&1) d-> name##flg|=EF_ON;				\
+				if (pth. name##flg&2) d-> name##flg|=EF_SUSTAIN;		\
+				if (pth. name##flg&4) d-> name##flg|=EF_LOOP;			\
+				d-> name##susbeg=d-> name##susend=pth. name##sus;		\
+				d-> name##beg=pth. name##beg;							\
+				d-> name##end=pth. name##end;							\
+				d-> name##pts=pth. name##pts;							\
+																		\
+				/* scale envelope */									\
+				for (p=0;p<XMENVCNT/2;p++)								\
+					d-> name##env[p].val<<=2;							\
+																		\
+				if ((d-> name##flg&EF_ON)&&(d-> name##pts<2))			\
+					d-> name##flg&=~EF_ON
+#else
+#define XM_ProcessEnvelope(name) 											\
+				for (u = 0; u < (XMENVCNT >> 1); u++) {						\
+					d-> name/**/env[u].pos = pth. name/**/env[u << 1];		\
+					d-> name/**/env[u].val = pth. name/**/env[(u << 1)+ 1];	\
+				}															\
+				if (pth. name/**/flg&1) d-> name/**/flg|=EF_ON;				\
+				if (pth. name/**/flg&2) d-> name/**/flg|=EF_SUSTAIN;		\
+				if (pth. name/**/flg&4) d-> name/**/flg|=EF_LOOP;			\
+				d-> name/**/susbeg=d-> name/**/susend=						\
+				                      pth. name/**/sus;						\
+				d-> name/**/beg=pth. name/**/beg;							\
+				d-> name/**/end=pth. name/**/end;							\
+				d-> name/**/pts=pth. name/**/pts;							\
+																			\
+				/* scale envelope */										\
+				for (p=0;p<XMENVCNT/2;p++)									\
+					d-> name/**/env[p].val<<=2;								\
+																			\
+				if ((d-> name/**/flg&EF_ON)&&(d-> name/**/pts<2))			\
+					d-> name/**/flg&=~EF_ON
+#endif			
 
-				if((d->volflg & EF_ON)&&(d->volpts<2))
-					d->volflg &= ~EF_ON;
+				XM_ProcessEnvelope(vol);
+				XM_ProcessEnvelope(pan);
+#undef XM_ProcessEnvelope
 
-				memcpy(d->panenv,pth.panenv,XMENVCNT);
-				d->panflg    = pth.panflg;
-				d->pansusbeg = d->pansusend = pth.pansus;
-				d->panbeg    = pth.panbeg;
-				d->panend    = pth.panend;
-				d->panpts    = pth.panpts;
-
-				/* scale panning envelope: */
-				for(p=0;p<XMENVCNT/2;p++)
-					d->panenv[p].val<<=2;
-				if((d->panflg & EF_ON) && (d->panpts<2))
-					d->panflg &= ~EF_ON;
+				if (d->volflg & EF_ON)
+					FixEnvelope(d->volenv, d->volpts);
+				if (d->panflg & EF_ON)
+					FixEnvelope(d->panenv, d->panpts);
 
 				/* Samples are stored outside the instrument struct now, so we
 				   have to load them all into a temp area, count the of.numsmp
@@ -541,6 +627,13 @@ static BOOL LoadInstruments(void)
 					of.numsmp+=ih.numsmp;
 			} else {
 				/* read the remainder of the header */
+				ck = _mm_ftell(modreader);
+				_mm_fseek(modreader,0,SEEK_END);
+				if ((headend<0) || (_mm_ftell(modreader)<headend) || (headend<ck)) {
+				_mm_fseek(modreader,ck,SEEK_SET);
+				break;
+				}
+				_mm_fseek(modreader,ck,SEEK_SET);
 				for(u=headend-_mm_ftell(modreader);u;u--) _mm_read_UBYTE(modreader);
 
 				if(_mm_eof(modreader)) {
@@ -626,8 +719,10 @@ BOOL XM_Load(BOOL curious)
 	of.numpos    = mh->songlength;               /* copy the songlength */
 	of.reppos    = mh->restart<mh->songlength?mh->restart:0;
 	of.numins    = mh->numins;
-	of.flags    |= UF_XMPERIODS|UF_INST|UF_BGSLIDES|UF_NOWRAP;
+	of.flags    |= UF_XMPERIODS | UF_INST | UF_NOWRAP | UF_FT2QUIRKS |
+				   UF_PANNING;
 	if(mh->flags&1) of.flags |= UF_LINEAR;
+	of.bpmlimit  = 32;
 
 	memset(of.chanvol,64,of.numchn);             /* store channel volumes */
 
@@ -685,11 +780,10 @@ BOOL XM_Load(BOOL curious)
 			q->loopend   >>= 1;
 		}
 
-		q->flags|=SF_OWNPAN;
+		q->flags|=SF_OWNPAN|SF_DELTA|SF_SIGNED;
 		if(s->type&0x3) q->flags|=SF_LOOP;
 		if(s->type&0x2) q->flags|=SF_BIDI;
 		if(s->type&0x10) q->flags|=SF_16BITS;
-		q->flags|=SF_DELTA|SF_SIGNED;
 	}
 
 	d=of.instruments;
@@ -721,7 +815,7 @@ CHAR *XM_LoadTitle(void)
 
 /*========== Loader information */
 
-MLOADER load_xm={
+MIKMODAPI MLOADER load_xm={
 	NULL,
 	"XM",
 	"XM (FastTracker 2)",
